﻿using System;
using System.Collections.Generic;
using System.Data.SQLite;
using System.Windows.Forms;

namespace RCM {
    [Serializable]
    public class SQLite : IDataSource {
        public string Title { get; set; }
        public HashSet<Guid> ExpandedNodes { get; } = new HashSet<Guid>();

        public string FileName { get; set; }
        public string AESPassword { get; set; }

        [NonSerialized]
        private SQLiteConnection _connection;
        [NonSerialized]
        private SQLiteControl _control;

        public override string ToString() => Title;

        private SQLiteConnection CreateConnection(string fileName, bool failIfMissing) {
            return new SQLiteConnection(string.Format("Data Source={0};Version=3;FailIfMissing={1}", fileName, failIfMissing ? "True" : "False"));
        }

        private void OpenConnection() {
            if (_connection == null) {
                _connection = CreateConnection(FileName, true);
            }
            _connection.Open();
        }

        public Dictionary<Guid, IRecord> SelectAll() {
            Dictionary<Guid, IRecord> records = new Dictionary<Guid, IRecord>();
            OpenConnection();
            using (SQLiteCommand command = new SQLiteCommand("SELECT \"data\", LENGTH(\"data\") AS \"datalength\" FROM \"data\"", _connection))
            using (SQLiteDataReader reader = command.ExecuteReader()) {
                while (reader.Read()) {
                    int dataBytesLength = reader.GetInt32(1);
                    byte[] dataBytes = new byte[dataBytesLength];
                    reader.GetBytes(0, 0, dataBytes, 0, dataBytesLength);
                    IRecord record = (IRecord)ConfigSerializer.DecryptObject(dataBytes, AESPassword);
                    records.Add(record.Id, record);
                }
            }
            _connection.Close();
            return records;
        }

        public void Insert(IRecord record) {
            OpenConnection();
            EnableForeignKeys();
            using (SQLiteCommand command = new SQLiteCommand("INSERT INTO \"data\" (\"id\", \"group\", \"data\") VALUES (@id, @group, @data)", _connection)) {
                command.Parameters.AddWithValue("@id", record.Id.ToByteArray());
                if (record.GroupId == Guid.Empty) {
                    command.Parameters.AddWithValue("@group", null);
                } else {
                    command.Parameters.AddWithValue("@group", record.GroupId.ToByteArray());
                }
                command.Parameters.AddWithValue("@data", ConfigSerializer.EncryptObject(record, AESPassword));
                command.ExecuteNonQuery();
            }
            _connection.Close();
        }

        public void Update(IRecord record) {
            OpenConnection();
            EnableForeignKeys();
            using (SQLiteCommand command = new SQLiteCommand("UPDATE \"data\" SET \"group\" = @group, \"data\" = @data WHERE \"id\" = @id", _connection)) {
                if (record.GroupId == Guid.Empty) {
                    command.Parameters.AddWithValue("@group", null);
                } else {
                    command.Parameters.AddWithValue("@group", record.GroupId.ToByteArray());
                }
                command.Parameters.AddWithValue("@data", ConfigSerializer.EncryptObject(record, AESPassword));
                command.Parameters.AddWithValue("@id", record.Id);
                command.ExecuteNonQuery();
            }
            _connection.Close();
        }

        public void Delete(IRecord record) {
            OpenConnection();
            EnableForeignKeys();
            using (SQLiteCommand command = new SQLiteCommand("DELETE FROM \"data\" WHERE \"id\" = @id", _connection)) {
                command.Parameters.AddWithValue("@id", record.Id.ToByteArray());
                command.ExecuteNonQuery();
            }
            _connection.Close();
        }

        private void EnableForeignKeys() {
            using (SQLiteCommand command = new SQLiteCommand("PRAGMA foreign_keys = ON", _connection)) {
                command.ExecuteNonQuery();
            }
        }

        public UserControl GetControl() {
            if (_control == null || _control.IsDisposed) {
                _control = new SQLiteControl(this);
            }
            return _control;
        }

        public void UpdateFromControl() {
            FileName = _control.textBoxDatabase.Text;
            AESPassword = _control.textBoxEncPassword.Text;
            _connection = null;
        }

        public void CreateSchema(string fileName) {
            using (SQLiteConnection tempConnection = CreateConnection(fileName, false)) {
                tempConnection.Open();
                SQLiteCommand command = new SQLiteCommand(@"
DROP TABLE IF EXISTS ""data"";
CREATE TABLE ""data"" (
    ""id"" binary(16) NOT NULL,
    ""group"" binary(16),
    ""data"" blob NOT NULL,
    PRIMARY KEY (""id""),
    FOREIGN KEY (""group"") REFERENCES ""data""(""id"") ON DELETE CASCADE ON UPDATE CASCADE
);", tempConnection);
                command.ExecuteNonQuery();
                tempConnection.Close();
            }
        }

        public string TestConnection(string fileName) {
            using (SQLiteConnection tempConnection = CreateConnection(fileName, true)) {
                tempConnection.Open();
                using (SQLiteCommand command = new SQLiteCommand("PRAGMA quick_check", tempConnection)) {
                    return (string)command.ExecuteScalar();
                }
            }
        }
    }
}
